import base64
from lavague.core.base_driver import BaseDriver
from websockets.server import serve
import asyncio
import websockets
import threading
from typing import Any, Optional, Mapping
import json
import uuid
from lavague.core.agents import WebAgent


class DriverServer(BaseDriver):
    prompt_agent: WebAgent
    port: int

    async def handler(self, websocket, path):
        if self.client is not None:
            await websocket.close()
            return
        self.client = websocket
        try:
            async for message in websocket:
                self.handle_client_response(message)
        except websockets.exceptions.ConnectionClosed as e:
            print("Client disconnected")
        finally:
            self.client = None

    def handle_prompt_agent_action(self, type: str, args: str):
        if self.prompt_agent is None:
            raise Exception(
                "No prompt agent set, attach it with driver.accept_prompts(agent)"
            )
        if type == "run":
            self.prompt_agent.run(args)
        elif type == "get":
            self.prompt_agent.get(args)

    def exec_prompt_agent_task(self, message: str):
        try:
            obj = json.loads(message)

            if "type" in obj:

                def target():
                    self.handle_prompt_agent_action(obj["type"], obj["args"])

                task = threading.Thread(target=target)
                task.start()
        except Exception:
            pass

    def handle_client_response(self, message):
        if message == "PING":
            return
        self.client_response = message
        self.exec_prompt_agent_task(message)

    def start_server(self):
        asyncio.set_event_loop(asyncio.new_event_loop())
        self.server = websockets.serve(
            self.handler, "localhost", self.port, max_size=20 * 1024 * 1024
        )
        asyncio.get_event_loop().run_until_complete(self.server)
        asyncio.get_event_loop().run_forever()

    def run_server_in_thread(self):
        server_thread = threading.Thread(target=self.start_server)
        server_thread.start()
        return server_thread

    async def send_message_to_client(self, message, id):
        if self.client is not None:
            self.client_response = None
            await self.client.send(message)
            while self.client_response is None:
                await asyncio.sleep(0.1)
            return self.client_response
        return None

    def send_command_and_get_response_sync(self, command, args=""):
        id = str(uuid.uuid4())
        event = threading.Event()
        response_data = {}

        data = {"command": command, "args": args, "id": id}

        json_string = json.dumps(data)

        def send_command():
            response = asyncio.run(self.send_message_to_client(json_string, id))
            response_data["response"] = response
            event.set()

        response_thread = threading.Thread(target=send_command)
        response_thread.start()

        # Wait for the response
        event.wait()

        ret = ""
        if "response" in response_data:
            try:
                jso = json.loads(response_data["response"])
                ret = jso["ret"]
                response_data.clear()
            except Exception as e:
                print(e)
                pass
        return ret

    def __init__(self, url: Optional[str] = None, port: int = 8000):
        self.client = None
        self.client_response = None
        self.server = None
        self.port = port
        self.thread = self.run_server_in_thread()
        while self.client is None:
            pass
        print("Connected")
        super().__init__(url, None)

    def default_init_code(self) -> Any:
        return None

    def code_for_init(self) -> str:
        return ""

    def get_html(self) -> str:
        html = self.send_command_and_get_response_sync("get_html")
        return html

    def get_driver(self) -> BaseDriver:
        return self

    def resize_driver(self, width, height) -> None:
        pass

    def code_for_resize(self, width, height) -> str:
        ""

    def get_url(self) -> Optional[str]:
        url = self.send_command_and_get_response_sync("get_url")
        return url

    def code_for_get(self, url: str) -> str:
        return ""

    def get(self, url: str) -> None:
        res = self.send_command_and_get_response_sync("get", url)

    def back(self) -> None:
        res = self.send_command_and_get_response_sync("back")

    def code_for_back(self) -> None:
        return ""

    def get_screenshot_as_png(self) -> bytes:
        scr = self.send_command_and_get_response_sync("get_screenshot")
        scr_bytes = base64.b64decode(scr.split(",")[1])
        return scr_bytes

    def destroy(self) -> None:
        self.server.ws_server.close()

    def accept_prompts(self, prompt_agent: WebAgent) -> None:
        self.prompt_agent = prompt_agent
        self.send_command_and_get_response_sync("accept_prompts")

    def check_visibility(self, xpath: str) -> bool:
        return self.send_command_and_get_response_sync("is_visible", xpath)
        # try:
        #     return self.driver.find_element(By.XPATH, xpath).is_displayed()
        # except:
        #     return False

    def get_highlighted_element(self, generated_code: str):
        # local_scope = {"driver": self.get_driver()}
        # assignment_code = keep_assignments(generated_code)
        # self.exec_code(assignment_code, locals=local_scope)

        # # We extract pairs of variables assigned during execution with their name and pointer
        # variable_names = return_assigned_variables(generated_code)

        # elements = []

        # for variable_name in variable_names:
        #     var = local_scope[variable_name]
        #     if type(var) == WebElement:
        #         elements.append(var)

        # if len(elements) == 0:
        #     raise ValueError(f"No element found.")

        # outputs = []
        # for element in elements:
        #     element: WebElement

        #     bounding_box = {}
        #     viewport_size = {}

        #     self.execute_script(
        #         "arguments[0].setAttribute('style', arguments[1]);",
        #         element,
        #         "border: 2px solid red;",
        #     )
        #     self.execute_script(
        #         "arguments[0].scrollIntoView({block: 'center'});", element
        #     )
        #     screenshot = self.get_screenshot_as_png()

        #     bounding_box["x1"] = element.location["x"]
        #     bounding_box["y1"] = element.location["y"]
        #     bounding_box["x2"] = bounding_box["x1"] + element.size["width"]
        #     bounding_box["y2"] = bounding_box["y1"] + element.size["height"]

        #     viewport_size["width"] = self.execute_script("return window.innerWidth;")
        #     viewport_size["height"] = self.execute_script("return window.innerHeight;")
        #     screenshot = BytesIO(screenshot)
        #     screenshot = Image.open(screenshot)
        #     output = {
        #         "screenshot": screenshot,
        #         "bounding_box": bounding_box,
        #         "viewport_size": viewport_size,
        #     }
        #     outputs.append(output)
        return None

    def exec_code(
        self,
        code: str,
        globals: dict[str, Any] = None,
        locals: Mapping[str, object] = None,
    ):
        if code is not None and len(code.strip()) > 0:
            url = self.send_command_and_get_response_sync("exec_code", code)
            return url
        else:
            return ""

    def execute_script(self, js_code: str, *args) -> Any:
        return self.send_command_and_get_response_sync("execute_script", js_code)

    def is_bottom_of_page(self) -> bool:
        ret = super().is_bottom_of_page()
        return ret["value"]

    def code_for_execute_script(self, js_code: str, *args) -> str:
        return (
            f"driver.execute_script({js_code}, {', '.join(str(arg) for arg in args)})"
        )

    def wait(self, time_between_actions):
        json_str = f"""[
    {{
        "action": {{
            "name": "wait",
            "args": {{
                "value": {time_between_actions}
            }}
        }}
    }}
]"""
        self.exec_code(json_str)
        pass

    def scroll_up(self):
        json_str = """[
    {
        "action": {
            "name": "scroll",
            "args": {
            "value": "up"
            }
        }
    }
]"""
        self.exec_code(json_str)
        pass

    def scroll_down(self):
        json_str = """[
    {
        "action": {
            "name": "scroll",
            "args": {
            "value": "down"
            }
        }
    }
]"""
        self.exec_code(json_str)
        pass

    def resize_driver(self, width, targeted_height):
        pass

    def get_capability(self) -> str:
        return REMOTE_PROMPT_TEMPLATE


REMOTE_PROMPT_TEMPLATE = """
You are a chrome extension and your goal is to interact with web pages. You have been given a series of HTML snippets and queries.
Your goal is to return a list of actions that should be done in order to execute the actions.
Always target elements by XPATH.

The actions available are: 
Name: The message port closed before a response was received.
Description: Click on an element with a specific xpath
Arguments:
  - xpath (string)

Name: setValue
Description: Focus on and set the value of an input element with a specific xpath
Arguments:
  - xpath (string)
  - value (string)

Name: setValueAndEnter
Description: Like "setValue", except then it presses ENTER. Use this tool can submit the form when there's no "submit" button.
Arguments:
  - xpath (string)
  - value (string)

Name: fail
Description: Indicate that you are unable to complete the task
No arguments.

Here are examples of previous answers:
HTML:
<div class="QS5gu ud1jmf" role="none" xpath="/html/body/div[2]/div[2]/div[3]/span/div/div/div/div[1]/div/div/button/div">Inloggen</div></button></div></div></div><div class="GZ7xNe" xpath="/html/body/div[2]/div[2]/div[3]/span/div/div/div/div[2]"><h1 class="I90TVb" id="S3BnEe" xpath="/html/body/div[2]/div[2]/div[3]/span/div/div/div/div[2]/h1">Voordat je verdergaat naar Google</h1><div class="AG96lb" xpath="/html/body/div[2]/div[2]/div[3]/span/div/div/div/div[2]/div"><div class="eLZYyf" xpath="/html/body/div[2]/div[2]/div[3]/span/div/div/div/div[2]/div/div[1]">We gebruiken <a class="F4a1l" href="https://policies.google.com/technologies/cookies?utm_source=ucbs&amp;hl=nl" target="_blank" xpath="/html/body/div[2]/div[2]/div[3]/span/div/div/div/div[2]/div/div[1]/a">cookies</a> en gegevens voor het volgende:<ul class="dbXO9" xpath="/html/body/div[2]/div[2]/div[3]/span/div/div/div/div[2]/div/div[1]/ul"><li class="gowsYd ibCF0c" xpath="/html/body/div[2]/div[2]/div[3]/span/div/div/div/div[2]/div/div[1]/ul/li[1]">Google-services leveren en onderhouden</li><li class="gowsYd GwwhGf" xpath="/html/body/div[2]/div[2]/div[3]/span/div/div/div/div[2]/div/div[1]/ul/li[2]">Uitval bijhouden en bescherming bieden tegen spam, fraude en misbruik</li><li class="gowsYd v8Bpfb" xpath="/html/body/div[2]/div[2]/div[3]/span/div/div/div/div[2]/div/div[1]/ul/li[3]">Doelgroepbetrokkenheid en sitestatistieken meten om inzicht te krijgen in hoe onze services worden gebruikt en de kwaliteit van die services te verbeteren</li></ul></div><div class="eLZYyf" xpath="/html/body/div[2]/div[2]/div[3]/span/div/div/div/div[2]/div/div[2]">Als je Alles accepteren kiest, gebruiken we cookies en gegevens ook voor het volgende:<ul class="dbXO9" xpath="/html/body/div[2]/div[2]/div[3]/span/div/div/div/div[2]/div/div[2]/ul"><li class="gowsYd M6j9qf" xpath="/html/body/div[2]/div[2]/div[3]/span/div/div/div/div[2]/div/div[2]/ul/li[1]">Nieuwe services ontwikkelen en verbeteren</li><li class="gowsYd v8Bpfb" xpath="/html/body/div[2]/div[2]/div[3]/span/div/div/div/div[2]/div/div[2]/ul/li[2]">Advertenties laten zien en de effectiviteit ervan meten</li><li class="gowsYd e21Mac" xpath="/html/body/div[2]/div[2]/div[3]/span/div/div/div/div[2]/div/div[2]/ul/li[3]">Gepersonaliseerde content laten zien (afhankelijk van je instellingen)</li><li class="gowsYd ohEWPc" xpath="/html/body/div[2]/div[2]/div[3]/span/div/div/div/div[2]/div/div[2]/ul/li[4]">Gepersonaliseerde advertenties laten zien (afhankelijk van je instellingen)</li></ul><div class="jLhwdc" xpath="/html/body/div[2]/div[2]/div[3]/span/div/div/div/div[2]/div/div[2]/div">Als je Alles afwijzen kiest, gebruiken we cookies niet voor deze aanvullende doeleinden.</div></div><div class="yS1nld" xpath="/html/body/div[2]/div[2]/div[3]/span/div/div/div/div[2]/div/div[3]">Niet-gepersonaliseerde content wordt beïnvloed door factoren zoals de content die je op dat moment bekijkt, activiteit in je actieve zoeksessie en je locatie. Niet-gepersonaliseerde advertenties worden beïnvloed door de content die je op dat moment bekijkt en je algemene locatie. Gepersonaliseerde content en advertenties kunnen ook relevantere resultaten, aanbevelingen en op jou toegespitste advertenties omvatten die zijn gebaseerd op eerdere activiteit van deze browser, zoals uitgevoerde Google-zoekopdrachten. We gebruiken cookies en gegevens ook om te zorgen dat de functionaliteit geschikt is voor je leeftijd, als dit relevant is.</div><div class="yS1nld" xpath="/html/body/div[2]/div[2]/div[3]/span/div/div/div/div[2]/div/div[4]">Selecteer Meer opties om meer informatie te bekijken, waaronder over hoe je je privacyinstellingen beheert. Je kunt ook altijd naar <span xpath="/html/body/div[2]/div[2]/div[3]/span/div/div/div/div[2]/div/div[4]/span">g.co/privacytools</span> gaan.</div></div></div><div class="spoKVd" xpath="/html/body/div[2]/div[2]/div[3]/span/div/div/div/div[3]">
<div class="spoKVd" xpath="/html/body/div[2]/div[2]/div[3]/span/div/div/div/div[3]"><div class="GzLjMd" xpath="/html/body/div[2]/div[2]/div[3]/span/div/div/div/div[3]/div[1]"><button class="tHlp8d" data-ved="0ahUKEwjX3bmBmKeGAxU2xQIHHcGoAg4Q4cIICHw" id="W0wltc" xpath="/html/body/div[2]/div[2]/div[3]/span/div/div/div/div[3]/div[1]/button[1]"><div class="QS5gu sy4vM" role="none" xpath="/html/body/div[2]/div[2]/div[3]/span/div/div/div/div[3]/div[1]/button[1]/div">Alles afwijzen</div></button><button class="tHlp8d" data-ved="0ahUKEwjX3bmBmKeGAxU2xQIHHcGoAg4QiZAHCH0" id="L2AGLb" xpath="/html/body/div[2]/div[2]/div[3]/span/div/div/div/div[3]/div[1]/button[2]"><div class="QS5gu sy4vM" role="none" xpath="/html/body/div[2]/div[2]/div[3]/span/div/div/div/div[3]/div[1]/button[2]/div">Alles accepteren</div></button></div><div class="GzLjMd" xpath="/html/body/div[2]/div[2]/div[3]/span/div/div/div/div[3]/div[2]"><button class="tHlp8d" data-ved="0ahUKEwjX3bmBmKeGAxU2xQIHHcGoAg4QiJAHCH4" id="VnjCcb" role="link" tabindex="-1" xpath="/html/body/div[2]/div[2]/div[3]/span/div/div/div/div[3]/div[2]/button"><a class="eOjPIe" tabindex="0" xpath="/html/body/div[2]/div[2]/div[3]/span/div/div/div/div[3]/div[2]/button/a">Meer opties</a></button></div></div><div class="XWlrff cG0Dmf" xpath="/html/body/div[2]/div[2]/div[3]/span/div/div/div/div[4]"><a class="peRL2e" data-ved="0ahUKEwjX3bmBmKeGAxU2xQIHHcGoAg4Qj5AHCH8" href="https://policies.google.com/privacy?hl=nl&amp;fg=1&amp;utm_source=ucbs" id="RP3V5c" xpath="/html/body/div[2]/div[2]/div[3]/span/div/div/div/div[4]/a[1]">Privacy</a>
<div class="spoKVd" xpath="/html/body/div[2]/div[2]/div[3]/span/div/div/div/div[3]"><div class="GzLjMd" xpath="/html/body/div[2]/div[2]/div[3]/span/div/div/div/div[3]/div[1]"><button class="tHlp8d" data-ved="0ahUKEwjX3bmBmKeGAxU2xQIHHcGoAg4Q4cIICHw" id="W0wltc" xpath="/html/body/div[2]/div[2]/div[3]/span/div/div/div/div[3]/div[1]/button[1]"><div class="QS5gu sy4vM" role="none" xpath="/html/body/div[2]/div[2]/div[3]/span/div/div/div/div[3]/div[1]/button[1]/div">Alles afwijzen</div></button><button class="tHlp8d" data-ved="0ahUKEwjX3bmBmKeGAxU2xQIHHcGoAg4QiZAHCH0" id="L2AGLb" xpath="/html/body/div[2]/div[2]/div[3]/span/div/div/div/div[3]/div[1]/button[2]"><div class="QS5gu sy4vM" role="none" xpath="/html/body/div[2]/div[2]/div[3]/span/div/div/div/div[3]/div[1]/button[2]/div">Alles accepteren</div></button></div><div class="GzLjMd" xpath="/html/body/div[2]/div[2]/div[3]/span/div/div/div/div[3]/div[2]"><button class="tHlp8d" data-ved="0ahUKEwjX3bmBmKeGAxU2xQIHHcGoAg4QiJAHCH4" id="VnjCcb" role="link" tabindex="-1" xpath="/html/body/div[2]/div[2]/div[3]/span/div/div/div/div[3]/div[2]/button"><a class="eOjPIe" tabindex="0" xpath="/html/body/div[2]/div[2]/div[3]/span/div/div/div/div[3]/div[2]/button/a">Meer opties</a></button></div></div><div class="XWlrff cG0Dmf" xpath="/html/body/div[2]/div[2]/div[3]/span/div/div/div/div[4]"><a class="peRL2e" data-ved="0ahUKEwjX3bmBmKeGAxU2xQIHHcGoAg4Qj5AHCH8" href="https://policies.google.com/privacy?hl=nl&amp;fg=1&amp;utm_source=ucbs" id="RP3V5c" xpath="/html/body/div[2]/div[2]/div[3]/span/div/div/div/div[4]/a[1]">Privacy</a>
Query: Click on the button labeled 'Alles accepteren' to accept all cookies.
Completion:
[
    {
        "action": {
            "name": "click",
            "args": {
            "xpath": "/html/body/div[2]/div[2]/div[3]/span/div/div/div/div[3]/div[1]/button[2]"
            }
        }
    }
]
---
HTML:
<div class="devsite-top-logo-row-middle" xpath="/html/body/section/devsite-header/div/div[1]/div/div/div[2]">
<div class="devsite-header-upper-tabs" xpath="/html/body/section/devsite-header/div/div[1]/div/div/div[2]/div[1]">
<devsite-tabs class="upper-tabs devsite-overflow-menu--open" connected="" xpath="/html/body/section/devsite-header/div/div[1]/div/div/div[2]/div[1]/devsite-tabs">
<nav aria-label="Upper tabs" class="devsite-tabs-wrapper" xpath="/html/body/section/devsite-header/div/div[1]/div/div/div[2]/div[1]/devsite-tabs/nav">
<tab class="devsite-active" xpath="/html/body/section/devsite-header/div/div[1]/div/div/div[2]/div[1]/devsite-tabs/nav/tab[1]">
<a aria-label="Gemini API, selected" class="devsite-tabs-content gc-analytics-event" data-category="Site-Wide Custom Events" data-label="Tab: Gemini API" href="https://ai.google.dev/gemini-api" track-metadata-eventdetail="https://ai.google.dev/gemini-api" track-metadata-module="primary nav" track-metadata-position="nav - gemini api" track-name="gemini api" track-type="nav" xpath="/html/body/section/devsite-header/div/div[1]/div/div/div[2]/div[1]/devsite-tabs/nav/tab[1]/a">
    Gemini API
  
    </a>
</tab>
<tab class="devsite-overflow-tab" xpath="/html/body/section/devsite-header/div/div[1]/div/div/div[2]/div[1]/devsite-tabs/nav/tab[2]"><!-- -->
<a aria-label="Extended Navigation" class="devsite-icon devsite-icon-arrow-drop-down" href="#" style="border: 2px solid red;" xpath="/html/body/section/devsite-header/div/div[1]/div/div/div[2]/div[1]/devsite-tabs/nav/tab[2]/a"><!--?lit$8296333005$-->More</a>
<div class="devsite-tabs-overflow-menu" scrollbars="" xpath="/html/body/section/devsite-header/div/div[1]/div/div/div[2]/div[1]/devsite-tabs/nav/tab[2]/div">
<tab xpath="/html/body/section/devsite-header/div/div[1]/div/div/div[2]/div[1]/devsite-tabs/nav/tab[2]/div/tab[1]">
<a class="devsite-tabs-content gc-analytics-event" data-category="Site-Wide Custom Events" data-label="Tab: Gemma" href="https://ai.google.dev/gemma" track-metadata-eventdetail="https://ai.google.dev/gemma" track-metadata-module="primary nav" track-metadata-position="nav - gemma" track-name="gemma" track-type="nav" xpath="/html/body/section/devsite-header/div/div[1]/div/div/div[2]/div[1]/devsite-tabs/nav/tab[2]/div/tab[1]/a">
    Gemma
  
    </a>
</tab><tab xpath="/html/body/section/devsite-header/div/div[1]/div/div/div[2]/div[1]/devsite-tabs/nav/tab[2]/div/tab[2]">
<a class="devsite-tabs-content gc-analytics-event" data-category="Site-Wide Custom Events" data-label="Tab: Google AI Edge" href="https://ai.google.dev/edge" track-metadata-eventdetail="https://ai.google.dev/edge" track-metadata-module="primary nav" track-metadata-position="nav - google ai edge" track-name="google ai edge" track-type="nav" xpath="/html/body/section/devsite-header/div/div[1]/div/div/div[2]/div[1]/devsite-tabs/nav/tab[2]/div/tab[2]/a">
    Google AI Edge
  

Query: Click on "Gemma" under the "More" dropdown menu.
Completion:
[
    {
        "action": {
            "name": "click",
            "args": {
            "xpath": "/html/body/section/devsite-header/div/div[1]/div/div/div[2]/div[1]/devsite-tabs/nav/tab[2]/a"
            }
        }
    },
    {
        "action": {
            "name": "click",
            "args": {
            "xpath": "/html/body/section/devsite-header/div/div[1]/div/div/div[2]/div[1]/devsite-tabs/nav/tab[2]/div/tab[1]/a"
            }
        }
    }
]
Your response must always be in JSON format and must include object "action", which contains the string "name" of tool of choice, and necessary arguments ("args") if required by the tool.
"""
